/*
 * Copyright (c) 2005 Björn Steinbrink
 *
 */

#include <stdlib.h>
#include <unistd.h>
#include <errno.h>
#include <signal.h>
#include <stdint.h>
#include <stdio.h>
#include <syslog.h>
#include <pty.h>
#include <utmp.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/wait.h>
#include <sys/un.h>

#include "vserver.h"
#include "switch.h"
#include "context.h"
#include "context_cmd.h"
#include "namespace.h"

#include "commands.h"

#define BACKLOG_SIZE 3

#define NAME "vservd"
#define SOCKET_PATH "/tmp/vservd.sock"

int s1;

int init_socket()
{
	struct sockaddr_un local;
	int len;

	s1 = socket(AF_UNIX, SOCK_STREAM, 0);
	if (s1 == -1) {
		perror("Couldn't create socket");
		return 1;
	}

	local.sun_family = AF_UNIX;
	strcpy(local.sun_path, SOCKET_PATH);
	len = strlen(local.sun_path) + sizeof(local.sun_family);
	unlink(local.sun_path);

	if (bind(s1, (struct sockaddr*)&local, len) == -1) {
		perror("Couldn't bind socket");
		return 1;
	}

	if (listen(s1, BACKLOG_SIZE) == -1) {
		perror("Can't listen on socket");
		return 1;
	}

	return 0;
}

int init()
{
	if (init_socket())
		return 1;

	return 0;
}

int do_loop()
{
	int quit = 0;

	while(!quit) {
		int s2, done = 0;

		s2 = accept(s1, NULL ,0);
		if (s2 == -1) {
			syslog(LOG_NOTICE, "Failed to accept connection: %s", strerror(errno));
		}

		while (!done) {
			int n;
			fd_set fds;
			struct vservd_cmd cmd;
			struct msghdr msg;
			struct iovec iov;

			FD_ZERO(&fds);
			FD_SET(s2, &fds);

			select(s2+1, &fds, NULL, NULL, NULL);

			memset(&msg, 0, sizeof(msg));
			iov.iov_base = &cmd;
			iov.iov_len = sizeof(cmd);
			msg.msg_iov = &iov;
			msg.msg_iovlen = 1;

			n = recvmsg(s2, &msg, 0);

			if (n == 0) {
				close(s2);
				break;
			}

			if (n == -1) {
				syslog(LOG_NOTICE, "Read from socket failed: %s", strerror(errno));
				close(s2);
				break;
			}

			syslog(LOG_NOTICE, "Received %d with xid %d", cmd.type, cmd.xid);
			switch (cmd.type) {
				case CMD_CREATE:
					if (!fork()) {
						struct vcmd_ctx_flags_v0 vc_data;

						close(s1);
						close(s2);

						if (vserver(VCMD_ctx_create_v0, cmd.xid, NULL) == -1)
							exit(EXIT_FAILURE);
						vc_data.mask = VXF_STATE_SETUP;
						vc_data.flagword = 0;
						vserver(VCMD_set_cflags, cmd.xid, &vc_data);
						sleep(10);
						exit(EXIT_SUCCESS);
					}
					break;
				case CMD_ENTER:
					if (!(fork())) {
						int master, slave;
						char control[sizeof(struct cmsghdr)+10];
						struct cmsghdr *cmsg;
						close(s1);
						vserver(VCMD_enter_namespace, cmd.xid, NULL);
						chroot(".");
						chdir("/");
						vserver(VCMD_ctx_migrate, cmd.xid, NULL);
						openpty(&master, &slave, NULL, NULL, NULL);
						login_tty(slave);
						memset(&msg, 0, sizeof(msg));
						iov.iov_base = "MASTER";
						iov.iov_len = sizeof("MASTER");
						msg.msg_iov = &iov;
						msg.msg_iovlen = 1;
						msg.msg_control = control;
						msg.msg_controllen = sizeof(control);

						cmsg = CMSG_FIRSTHDR(&msg);
						cmsg->cmsg_level = SOL_SOCKET;
						cmsg->cmsg_type = SCM_RIGHTS;
						cmsg->cmsg_len = CMSG_LEN(sizeof(master));
						*(int *)CMSG_DATA(cmsg) = master;
						msg.msg_controllen = cmsg->cmsg_len;

						if (sendmsg(s2, &msg, 0) == -1) {
							syslog(LOG_NOTICE, "sendmsg");
							exit(EXIT_FAILURE);
						}
						close(s2);
						char *args[2];
						args[0] = "/bin/login";
						args[1] = NULL;
						execv(args[0], args);
					}
					wait(NULL);
					break;
				case CMD_QUIT:
					done = 1;
					quit = 1;
					break;
			}
		}
		close(s2);
	}

	return 0;
}

int main(int argc, char **argv)
{
	int ret = EXIT_SUCCESS;

	if (init())
		exit(EXIT_FAILURE);

	daemon(0, 0);
	openlog(NAME, LOG_ODELAY, LOG_DAEMON);
	syslog(LOG_INFO, "daemon has been started with pid %d", getpid());

	if (do_loop())
		ret = EXIT_FAILURE;

	close(s1);

	syslog(LOG_INFO, "finished");
	exit(ret);
}
